# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This module supports interacting with the Slycat server from Python and the command line.

__version__ = "4.5.0"

# standard libraries

# parse arguments
import argparse
import os
import shlex

# authenticate to server
import getpass
import json

# encode/decode functions for string data (for http communication)
import base64

# miscellaneous utilities
import sys
import time
import math

# stream to server
import io

# logging
import logging

# url/IP
from urllib.parse import urlparse
import socket

# 3rd party libraries

# http requests & kerberos
import requests
import requests.exceptions as exceptions
from requests_kerberos import HTTPKerberosAuth, OPTIONAL
from requests_kerberos.exceptions import KerberosExchangeError

# turn off warning for insecure connections
from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

# handle arrays for Slycat requests
import numpy

# web server interaction
import cherrypy

# local libraries

# handle arrays for Slycat requests
import slycat.darray

# set up logging
log = logging.getLogger("slycat.web.client")
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler())
log.handlers[0].setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
log.propagate = False

# default file slice size (10 MB)
FILE_SLICE_DEFAULT = 10_000_000

# default Slycat server
HOST_DEFAULT = "https://localhost"

# private function (denoted by _name)
def _require_array_ranges(ranges):
  """Validates a range object (hyperslice) for transmission to the server."""

  if ranges is None:
    return None
  elif isinstance(ranges, int):
    return [(0, ranges)]
  elif isinstance(ranges, tuple):
    return [ranges]
  elif isinstance(ranges, list):
    return ranges
  else:
    cherrypy.log.error("slycat.web.client.__init__.py", "Not a valid ranges object.")
    raise Exception("Not a valid ranges object.")

# print iterations progress bar -- from stack overflow
# https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
def _print_progress_bar (iteration, total, prefix = '', suffix = '', 
  decimals = 1, length = 100, fill = '█', printEnd = "\r"):
    
    # Call in a loop to create terminal progress bar
    # @params:
    #     iteration   - Required  : current iteration (Int)
    #     total       - Required  : total iterations (Int)
    #     prefix      - Optional  : prefix string (Str)
    #     suffix      - Optional  : suffix string (Str)
    #     decimals    - Optional  : positive number of decimals in percent complete (Int)
    #     length      - Optional  : character length of bar (Int)
    #     fill        - Optional  : bar fill character (Str)
    #     printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end = printEnd)

    # print new line on complete
    if iteration == total: 
        print()

class ArgumentParser(argparse.ArgumentParser):
  """Return an instance of argparse.ArgumentParser, pre-configured with arguments to
  connect to a Slycat server.

  Pre-configured options are: 

    * **-\\-host** -- the URL of the Slycat server.
    * **-\\-port** -- the port of the Slycat server.
    * **-\\-http-proxy** -- the URL for an http proxy.
    * **-\\-https-proxy** -- the URL for an https proxy.
    * **-\\-verify** -- an SSL certificate to verify https host.
    * **-\\-no-verify** -- disable https verification.
    * **-\\-file-slice-size** -- maximum number of bytes to upload at once.
    * **-\\-user** -- user name.
    * **-\\-passowrd** -- user password.
    * **-\\-kerberos** -- enable kerberos authentication.
    * **-\\-log-level** -- log detail to display.
  """

  def __init__(self, *arguments, **keywords):

    argparse.ArgumentParser.__init__(self, *arguments, **keywords)

    # arguments for connecting to Slycat server

    self.add_argument("--host", default=HOST_DEFAULT, 
      help="Root URL of the Slycat server.  Default: '%(default)s'.")
    self.add_argument("--port", default=None, 
      help="Port of the Slycat server.")

    self.add_argument("--http-proxy", default="", 
      help="HTTP proxy URL.  Default: '%(default)s'.")
    self.add_argument("--https-proxy", default="", 
      help="HTTPS proxy URL.  Default: '%(default)s'.")
    self.add_argument("--verify", default=None, 
      help="Specify a certificate to use for HTTPS host certificate verification.")
    self.add_argument("--no-verify", default=False, action="store_true", 
      help="Disable HTTPS host certificate verification.")

    self.add_argument('--file-slice-size', default=FILE_SLICE_DEFAULT, type=int,
      help="Maximum number of bytes to upload before slicing files (i.e. before "
           "uploading in chunks).  Default: %(default)s.")

    self.add_argument("--user", default=getpass.getuser(), 
      help="Slycat username.  Default: '%(default)s'")
    self.add_argument("--password", default=None, help="User password.")

    self.add_argument("--kerberos", default=False, action="store_true", 
      help="Use Kerberos authentication.  Default: %(default)s.")

    self.add_argument("--log-level", default="info", 
      choices=["debug", "info", "warning", "error", "critical"], 
      help="Log level.  Default: '%(default)s'.")

  def parse_args(self, list_input=None):
    """Overrides argparse parse_args command.  Parses slycat.web.client arguments
    in addition to any arguments added by users of the class.  Can also be used to
    parse a list of arguments passed from a Python function, such as
    
    >>> parser = slycat.web.client.ArgumentParser()
    >>> parser.parse_args(["--host", "slycat.sandia.gov", "--kerberos"]).
    """

    if "SLYCAT" in os.environ:
      sys.argv += shlex.split(os.environ["SLYCAT"])

    # parse arguments
    arguments = argparse.ArgumentParser.parse_args(self, list_input)

    # log level
    if arguments.log_level == "debug":
      log.setLevel(logging.DEBUG)
    elif arguments.log_level == "info":
      log.setLevel(logging.INFO)
    elif arguments.log_level == "warning":
      log.setLevel(logging.WARNING)
    elif arguments.log_level == "error":
      log.setLevel(logging.ERROR)
    elif arguments.log_level == "critical":
      log.setLevel(logging.CRITICAL)

    return arguments

class Connection(object):
  """Provides a class to facilitate communication with the Slycat web server.  
  To use, open a connection and submit requests.  For example:

  >>> parser = slycat.web.client.ArgumentParser()
  >>> arguments = parser.parse_args()
  >>> connection = slycat.web.client.connect(arguments)
  >>> projects = connection.get_projects()
  """

  def __init__(self, host=HOST_DEFAULT, port=None, kerberos=False,
               file_slice_size=FILE_SLICE_DEFAULT, **keywords):
    
    # proxies default to ""
    proxies = keywords.get("proxies", {"http": "", "https": ""})
    
    # certificate verification is disabled unless specifically requested
    verify = True
    if keywords.get("verify") == "False":
      verify = False
    elif not keywords.get("verify"):
      verify = False

    # resolve host alias
    host_alias = urlparse(host)
    host_alias = host_alias.netloc
    host_ip = socket.gethostbyname(host_alias)
    host_name = socket.gethostbyaddr(host_ip)
    host_name = host_name[0]

    # host and port
    self.host = host
    if host != HOST_DEFAULT:
      self.host = host.replace(host_alias, host_name)
    if port:
      self.host = self.host + ':' + port

    # using kerberos authentication?
    self.kerberos = kerberos

    # set up session
    self.keywords = keywords
    self.session = requests.Session()

    # set up max file slice size
    self.file_slice_size = file_slice_size

    # check for Kerberos authentication
    if self.kerberos:

      # get Kerberos ticket granting ticket
      self.session.auth = HTTPKerberosAuth(mutual_authentication=OPTIONAL,
                                           force_preemptive=True)

      # check that ticket is valid using list markings
      url = self.host + "/api/configuration/markings"
      try:
        response = self.session.get(url, proxies=proxies, verify=verify,
                                    auth=self.session.auth)

      except KerberosExchangeError:
        cherrypy.log.error("Could not find Kerberos ticket, try running 'kinit'.")
        sys.exit(1)

      if response.status_code == 401:
        cherrypy.log.error(
          "User %s is not Kerberos authenticated, try running 'kinit'." % 
          keywords.get("auth", ("",""))[0])
        sys.exit(1)

    else:

      # get user name and password, if supplied
      user_name, password = keywords.get("auth", ("",""))

      # if password not supplied, prompt user
      if not password:
        password = getpass.getpass("%s password: " % user_name)

      # encode as base 64
      user_name_b64 = base64.b64encode(user_name.encode("ascii"))
      password_b64 = base64.b64encode(password.encode("ascii"))

      # change back to ascii for JSON
      user_name_str = user_name_b64.decode('ascii')
      password_str = password_b64.decode('ascii')

      # authentication information
      data = {"user_name":user_name_str, "password":password_str}

      # login url
      url = self.host + "/api/login"

      # connect to server
      self.session.post(url, json=data, proxies=proxies, verify=verify)

      # session error, assume bad password
      if len(list(self.session.cookies.keys())) == 0:
        raise NameError('bad username or password:%s, for username:%s' % 
          (user_name, password))

  def request(self, method, path, **keywords):
    """Basic request to Slycat server using open session.  To make a request 
    provide the HTTP method and path, and this method will return the body of
    the response.  Additional keyword arguments must be compatible with the
    Python requests library.
    
    Parameters
    ----------
    method: string, required
      The HTTP method to use, e.g. "GET", "PUT", etc.
    path: string, required
      The extension to the URL for the Slycat server, e.g. "/api/models/(mid)".

    Returns
    -------
    response: the body of the Server response.
    """

    # combine per-request and per-connection keyword arguments
    keywords.update(self.keywords)
    
    # combine host and path to produce the final request URI
    uri = self.host + path
    
    # create log message for this request
    log_message = "{} {} {}".format(keywords.get("auth", ("", ""))[0], method, uri)

    # send request
    try:
      response = self.session.request(method, uri, **keywords)

      # add response to log message
      log_message += " => {} {}".format(response.status_code, response.raw.reason)

      # check to see if an error occured
      response.raise_for_status()

      # parse and return body of response
      body = None
      if response.headers["content-type"].startswith("application/json"):
        body = response.json()
      else:
        body = response.content

      # show request in debug log
      log.debug(log_message)

      return body

    # log any errors to both slycat.web.client and cherrypy
    except:
      log.debug(log_message)
      cherrypy.log.error("slycat.web.client.__init__.py request", "%s" % log_message)
      raise

  #######################################################################
  # Low-level functions that map directly to the underlying RESTful API #
  #######################################################################

  def delete_model(self, mid):
    """Delete an existing model.

    Parameters
    ----------
    mid: string, required
      The unique model identifier.

    See Also
    --------
    :http:delete:`/api/models/(mid)`
    """

    self.request("DELETE", "/api/models/%s" % (mid))

  def delete_project(self, pid):
    """Delete an existing project.

    Parameters
    ----------
    pid: string, required
      The unique project identifier.

    See Also
    --------
    :http:delete:`/api/projects/(pid)`
    """

    self.request("DELETE", "/api/projects/%s" % (pid))

  def delete_project_cache_object(self, pid, key):
    """Delete an existing project cache object.

    Parameters
    ----------
    pid: string, required
      The unique project identifier.
    key: string, required
      Unique cache object key.

    See Also
    --------
    :http:delete:`/api/projects/(pid)/cache/(key)`
    """

    self.request("DELETE", "/api/projects/%s/cache/%s" % (pid, key))

  def delete_reference(self, rid):
    """Delete an existing reference.

    Parameters
    ----------
    rid: string, required
      The unique reference identifier.

    See Also
    --------
    :http:delete:`/api/references/(rid)`
    """

    self.request("DELETE", "/api/references/%s" % (rid))

  def delete_remote(self, sid):
    """Delete an existing remote session.

    Parameters
    ----------
    sid: string, required
      The unique remote session identifier.

    See Also
    --------
    :http:delete:`/api/remotes/(hostname)`
    """

    self.request("DELETE", "/api/remotes/%s" % (sid))

  def get_bookmark(self, bid):
    """Retrieve an existing bookmark.

    Parameters
    ----------
    bid: string, required
      The unique bookmark identifier.

    Returns
    -------
    bookmark: object
      The bookmark object, which is an arbitrary collection of
      JSON-compatible data.

    See Also
    --------
    :http:get:`/api/bookmarks/(bid)`
    """

    return self.request("GET", "/api/bookmarks/%s" % (bid))

  def get_configuration_markings(self):
    """Retrieve marking information from the server.

    Returns
    -------
    markings: server marking information.

    See Also
    --------
    :http:get:`/api/configuration/markings`
    """

    return self.request("GET", "/api/configuration/markings", 
      headers={"accept":"application/json"})

  def get_configuration_parsers(self):
    """Retrieve parser plugin information from the server.

    Returns
    -------
    parsers: server parser plugin information.

    See Also
    --------
    :http:get:`/api/configuration/parsers`
    """

    return self.request("GET", "/api/configuration/parsers", 
      headers={"accept":"application/json"})

  def get_configuration_remote_hosts(self):
    """Retrieve remote host information from the server.

    Returns
    -------
    remote_hosts: server remote host information.

    See Also
    --------
    :http:get:`/api/configuration/remote-hosts`
    """

    return self.request("GET", "/api/configuration/remote-hosts", 
      headers={"accept":"application/json"})

  def get_configuration_support_email(self):
    """Retrieve support email information from the server.

    Returns
    -------
    email: server support email information.

    See Also
    --------
    :http:get:`/api/configuration/support-email`
    """

    return self.request("GET", "/api/configuration/support-email", 
      headers={"accept":"application/json"})

  def get_configuration_version(self):
    """Retrieve version information from the server.

    Returns
    -------
    version: server version information.

    See Also
    --------
    :http:get:`/api/configuration/version`
    """

    return self.request("GET", "/api/configuration/version",
      headers={"accept":"application/json"})

  def get_configuration_wizards(self):
    """Retrieve wizard plugin information from the server.

    Returns
    -------
    wizards: server wizard plugin information.

    See Also
    --------
    :http:get:`/api/configuration/wizards`
    """

    return self.request("GET", "/api/configuration/wizards", 
      headers={"accept":"application/json"})

  # the following three calls seem to be defunct

  # def get_global_resource(self, resource):
  #   return self.request("GET", "/resources/global/%s" % resource)

  # def get_model_resource(self, mtype, resource):
  #   return self.request("GET", "/resources/pages/%s/%s" % (mtype, resource))

  # def get_wizard_resource(self, wtype, resource):
  #   return self.request("GET", "/resources/wizards/%s/%s" % (wtype, resource))

  def get_model(self, mid):
    """Retrieve an existing model.

    Parameters
    ----------
    mid: string, required
      The unique model identifier

    Returns
    -------
    model: object
      The model object, which is an arbitrary collection of
      JSON-compatible data.

    See Also
    --------
    :http:get:`/api/models/(mid)`
    """

    return self.request("GET", "/api/models/%s" % mid, 
      headers={"accept":"application/json"})

  def get_model_arrayset_metadata(self, mid, aid, arrays=None, statistics=None, unique=None):
    """Retrieve metadata describing an existing model arrayset artifact.

    Parameters
    ----------
    mid: string, required
      The unique model identifier.
    aid: string, required
      The unique artifact identifier.
    arrays: string, optional
      A set of arrays, specified using HQL.
    statistics: string, optional
      A set of attributes, specified using HQL.
    unique: string, optional
      A set of attributes, specified using HQL.

    Returns
    -------
    metadata: object
      The arrayset metadata, which is an arbitrary collection of
      JSON-compatible data.

    See Also
    --------
    :http:get:`/api/models/(mid)/arraysets/(aid)/metadata`
    """

    params = dict()
    if arrays is not None:
      params["arrays"] = arrays
    if statistics is not None:
      params["statistics"] = statistics
    if unique is not None:
      params["unique"] = unique
    return self.request("GET", "/api/models/%s/arraysets/%s/metadata" % 
      (mid, aid), params=params, headers={"accept":"application/json"})

  def get_model_file(self, mid, aid):
    """Retrieves the file corresponding to a given model and artifacts.

    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    aid: string, required
      Unique (with the model) artifact id.

    Returns
    -------
    file: File data corresponding to artifact.

    See Also
    --------
    :http:get:`/api/models/(mid)/files/(aid)`
    """

    return self.request("GET", "/api/models/%s/files/%s" % (mid, aid))

  def get_model_parameter(self, mid, aid):
    """Retrieve a model parameter artifact.

    Model parameters are JSON objects of arbitrary complexity.  They are stored directly within the model
    as part of its database record, so they should be limited in size (larger data should be stored using
    arraysets or files).

    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    aid: string, required
      Unique (within the model) artifact id.

    Returns
    -------
    parameter: JSON-compatible object.

    See Also
    --------
    :http:put:`/api/models/(mid)/parameters/(aid)`
    """

    return self.request("GET", "/api/models/%s/parameters/%s" % 
      (mid, aid), headers={"accept":"application/json"})

  def get_project_models(self, pid):
    """Retrieve every model in a oroject.

    Parameters
    ----------
    pid: string, required
      Unique project identifier.

    Returns
    -------
    models: JSON-compatible object.

    See Also
    --------
    :http:get:`/api/projects/(pid)/models`
    """

    return self.request("GET", "/api/projects/%s/models" % pid, 
      headers={"accept":"application/json"})

  def get_project_references(self, pid):
    """Returns every reference in a project.
    
    Parameters
    ----------
    pid: string, required
      Unique project identifier.

    Returns
    -------
    references: JSON-compatible object.

    See Also
    --------
    :http:get:`/api/projects/(pid)/references`
    """
    
    return self.request("GET", "/api/projects/%s/references" % pid, 
      headers={"accept":"application/json"})

  def get_project(self, pid):
    """Retrieve an existing project.

    Parameters
    ----------
    pid: string, required
      Unique project identifier.

    Returns
    -------
    project: Arbitrary collection of JSON-compatible data.

    See Also
    --------
    :http:get:`/api/projects/(pid)`
    """

    return self.request("GET", "/api/projects/%s" % pid, 
      headers={"accept":"application/json"})

  def get_project_cache_object(self, pid, key):
    """Retrieve an object from a project cache.

    Parameters
    ----------
    pid: string, required
      Unique project identifier.
    key: string, required
      Cache object identifier.

    Returns
    -------
    content: Cached object content.

    See Also
    --------
    :http:get:`/api/projects/(pid)/cache/(key)`
    """

    return self.request("GET", "/api/projects/%s/cache/%s" % (pid, key))

  def get_projects(self):
    """Retrieve all projects.

    Returns
    -------
    projects: List of projects.  Each project is an arbitrary collection of JSON-compatible data.

    See Also
    --------
    :http:get:`/api/projects`
    """

    return self.request("GET", "/api/projects_list", headers={"accept":"application/json"})

  def get_remote_file(self, sid, path, cache=None, project=None, key=None):
    """Retrieve a file using a remote session.

    Parameters
    ----------
    sid: string, required
      Unique remote session identifier.
    path: string, required
      Remote filesystem path (must be absolute).
    cache: string, optional
      Optional server-side cache for the retrieved file.  Must be `None` or "project".
    project: string, optional
      If `cache` is set to "project", this must specify a unique project identifier.
    key: string, optional
      if `cache` is set to "project", this must specify a unique key for the cached object.

    Returns
    -------
    file: Remote file contents.

    See Also
    --------
    :http:get:`/api/remotes/(hostname)/file(path)`
    """

    return self.request("GET", "/api/remotes/%s/file%s" % (sid, path), 
      params={"cache": cache, "project": project, "key": key})

  def get_remote_image(self, sid, path, cache=None, project=None, key=None):
    """Retrieve an image using a remote session.

    Parameters
    ----------
    sid: string, required
      Unique remote session identifier.
    path: string, required
      Remote filesystem path (must be absolute).
    cache: string, optional
      Optional server-side cache for the retrieved image.  Must be `None` or "project".
    project: string, optional
      If `cache` is set to "project", this must specify a unique project identifier.
    key: string, optional
      if `cache` is set to "project", this must specify a unique key for the cached object.

    Returns
    -------
    image: Remote image contents.

    See Also
    --------
    :http:get:`/api/remotes/(hostname)/image(path)`
    """

    return self.request("GET", "/api/remotes/%s/image%s" % (sid, path), 
      params={"cache": cache, "project": project, "key": key})

  def get_user(self, uid=None):
    """Retrieve directory information about an existing user.

    Parameters
    ----------
    uid: string, optional
      Unique user identifier.  If unspecified, returns information about the user making the call.

    Returns
    -------
    user: Arbitrary collection of JSON-compatible data.

    See Also
    --------
    :http:get:`/api/users/(uid)`
    """

    return self.request("GET", "/api/users/%s" % ("-" if uid is None else uid), 
      headers={"accept":"application/json"})

  def post_events(self, path, parameters={}):
    """Post event to be logged on Slycat server.

    Parameters
    ----------
    path: string, required
      Path-like URI describing event to be logged.
    parameters: dictionary, optional
      JSON type object describing event.

    See Also
    --------
    :http:post:`/api/events/(event)`
    """

    self.request("POST", "/api/events/%s" % path, params=parameters)

  def post_model_files(self, mid, aids, files, parser, input=True, parameters={}):
    """Stores model file artifacts.
    
    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    aids: array, required
      Artifact IDs for model storage.
    files: array, required
      Local files for upload.
    parser: string, required
      Name of Slycat parser that will process the files.
    input: boolean, optional
      Set as true (default) to store as model artifacts.
    parametrs: dictionary, optional
      Additional data to pass to parser.

    See Also
    --------
    :http:post:`/api/models/(mid)/files`
    """

    data = parameters

    data.update({
      "input": json.dumps(input),
      "aids": aids,
      "parser": parser
    })

    files = [("files", ("blob", file)) for file in files]

    self.request("POST", "/api/models/%s/files" % mid, data=data, files=files)

  def post_model_finish(self, mid):
    """Notify the server that a model is fully initialized.

    When called, the server will perform one-time computation
    for the given model type.

    Parameters
    ----------
    mid: string, required
      Unique model identifier.

    See Also
    --------
    :http:post:`/api/models/(mid)/finish`
    """

    self.request("POST", "/api/models/%s/finish" % (mid))

  def post_project_bookmarks(self, pid, bookmark):
    """Store a bookmark.

    Parameters
    ----------
    pid: string, required
      Unique project identifier.
    bookmark: object
      Arbitrary collection of JSON-compatible data.

    Returns
    -------
    bid: string
      Unique bookmark identifier.

    See Also
    --------
    :http:post:`/api/projects/(pid)/bookmarks`
    """

    return self.request("POST", "/api/projects/%s/bookmarks" % (pid), 
      headers={"content-type":"application/json"}, data=json.dumps(bookmark))["id"]

  def post_project_models(self, pid, mtype, name, marking="", description=""):
    """Creates a new model, returning the model ID.
    
    Parameters
    ----------
    pid: string, required
      Unique project identifier.
    mtype: string, required
      Model type.
    name: string, required
      Model name.
    marking: string, optional
      Model marking.
    description: string, optional
      Description of model.

    Returns
    -------
    mid: string
      New model identifier.

    See Also
    --------
    :http:post:`/api/projects/(pid)/models`
    """

    return self.request("POST", "/api/projects/%s/models" % (pid), 
      headers={"content-type":"application/json"}, 
      data=json.dumps({"model-type":mtype, "name":name, "marking":marking, 
        "description":description}))["id"]

  def post_project_references(self, pid, name, mtype=None, mid=None, bid=None):
    """Store a project reference.

    Parameters
    ----------
    pid: string, required
      Unique project identifier.
    name: string, required
      Reference name.
    mtype: string, optional
      Optional model type.
    mid: string, optional
      Optional model identifier.
    bid: string, optional
      Optional bookmark identifier.

    Returns
    -------
    rid: string
      Unique reference identifier.

    """

    return self.request("POST", "/api/projects/%s/references" % (pid), 
      headers={"content-type":"application/json"}, data=json.dumps({"name":name, "model-type":mtype, "mid":mid, "bid":bid}))["id"]

  def post_projects(self, name, description=""):
    """Creates a new project, returning the project ID.
    
    Parameters
    ----------
    name: string, required
      Name of project to be created.
    description: string, optional
      Description of new project.

    Returns
    -------
    pid: string
      Unique project identifier.

    See Also
    --------
    :http:post:`/api/projects`
    """
    return self.request("POST", "/api/projects", 
      headers={"content-type":"application/json"}, 
      data=json.dumps({"name":name, "description":description}))["id"]

  def post_remotes(self, hostname, username, password, agent=None):
    """Creates a new remote connection from the Slycat server to another host. 

    Parameters
    ----------
    hostname: string, required
      Name of remote host.
    username: string, required
      User name for connection.
    password: string, required
      Password to authenticate connection.
    agent: boolean, optional
      Create an agent upon establishing connection.
    
    Returns
    -------
    sid: string
      Session ID for connection.

    See Also
    --------
    :http:post:`/api/remotes`
    """
    return self.request("POST", "/api/remotes", 
      headers={"content-type":"application/json"}, 
      data=json.dumps({"hostname":hostname, "username":username, 
                       "password":password, "agent": agent}))["sid"]

  def post_remote_browse(self, sid, path, file_reject=None, file_allow=None, 
                         directory_allow=None, directory_reject=None):
    """Uses an existing remote session to retrieve remote filesystem information.

    Parameters
    ----------
    sid: string, required
      Session ID for connection.
    path: string, required
      Remote file system path (must be absolute).
    file_reject: string, optional
      Regular expression for rejecting files.
    file_allow: string, optional
      Regular expression for retaining files.
    directory_reject: string, optional
      Regular expression for rejecting directories.
    directory_allow: string, optional
      Regular expression for retaining directories.

    Returns
    -------
    response_body: JSON like object.

    See Also
    --------
    :http:post:`/api/remotes/(hostname)/browse(path)`
    """

    body = {}
    if file_reject is not None:
      body["file-reject"] = file_reject
    if file_allow is not None:
      body["file-allow"] = file_allow
    if directory_reject is not None:
      body["directory-reject"] = directory_reject
    if directory_allow is not None:
      body["directory-allow"] = directory_allow

    return self.request("POST", "/api/remotes/" + sid + "/browse" + path, 
      headers={"content-type":"application/json"}, data=json.dumps(body))

  def put_model(self, mid, model):
    """Modify a Slycat model.

    Parameters
    ----------
    mid: string, required
      Model identifier.
    model: dictionary, required
      JSON like dictionary with fields to modified model including:

        * name (*string, optional*)
        * description (*string, optional*)
        * state (*string, optional*)
        * progress (*float, optional*)
        * message (*string, optional*)

    See Also
    --------
    :http:put:`/api/models/(mid)`
    """
    self.request("PUT", "/api/models/%s" % (mid), 
      headers={"content-type":"application/json"}, data=json.dumps(model))

  def put_model_arrayset_data(self, mid, aid, hyperchunks, data, force_json=False):
    """Write data to an arrayset artifact on the server.

    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    aid: string, required
      Unique (to the model) arrayset artifact id.
    hyperchunks: string, required
      Specifies where the data will be stored, in :ref:`Hyperchunks` format.
    data: iterable, required)
      A collection of numpy.ndarray data chunks to be uploaded.  The number of
      data chunks must match the number implied by the `hyperchunks` parameter.
    force_json: bool, optional)
      Force the client to upload data using JSON instead of the binary format.

    See Also
    --------
    :http:put:`/api/models/(mid)/arraysets/(aid)/data`
    """

    # Sanity check arguments
    if not isinstance(mid, str):
      cherrypy.log.error("slycat.web.client.__init__.py put_model_arrayset_data", 
        "Model id must be a string")
      raise ValueError("Model id must be a string.")
    if not isinstance(aid, str):
      cherrypy.log.error("slycat.web.client.__init__.py put_model_arrayset_data", 
        "Artifact id must be a string")
      raise ValueError("Artifact id must be a string.")
    if not isinstance(hyperchunks, str):
      cherrypy.log.error("slycat.web.client.__init__.py put_model_arrayset_data", 
        "Hyperchunks specification must be a string.")
      raise ValueError("Hyperchunks specification must be a string.")
    for chunk in data:
      if not isinstance(chunk, numpy.ndarray):
        cherrypy.log.error("slycat.web.client.__init__.py put_model_arrayset_data", 
          "Data chunk must be a numpy array.")
        raise ValueError("Data chunk must be a numpy array.")

    # Mark whether every data chunk is numeric ... if so, we can send the data in binary form.
    use_binary = numpy.all([chunk.dtype.char != "U" for chunk in data]) and not force_json

    # Build-up the request
    request_data = {}
    request_data["hyperchunks"] = hyperchunks
    if use_binary:
      request_data["byteorder"] = sys.byteorder

    if use_binary:

      # binary data
      request_buffer = io.BytesIO()
      for chunk in data:
        request_buffer.write(chunk.tostring(order="C"))
    else:

      # string data
      request_buffer = io.StringIO()
      request_buffer.write(json.dumps([chunk.tolist() for chunk in data]))

    # Send the request to the server ...
    self.request("PUT", "/api/models/%s/arraysets/%s/data" % (mid, aid), 
      data=request_data, files={"data":request_buffer.getvalue()})

  def put_model_arrayset_array(self, mid, aid, array, dimensions, attributes):
    """Starts a new array set array, ready to receive data.
    
    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    aid: string, required
      Unique artifact identifier.
    array: int, required
      Unique array index.
    dimensions: array, required
      Array dimensions.
    attributes: array, required
      Array attributes (data types).

    See Also
    --------
    :http:put:`/api/models/(mid)/arraysets/(aid)/arrays/(array)`
    """

    stub = slycat.darray.Stub(dimensions, attributes)
    self.request("PUT", "/api/models/%s/arraysets/%s/arrays/%s" % (mid, aid, array), 
      headers={"content-type":"application/json"}, 
      data=json.dumps({"dimensions":stub.dimensions, "attributes":stub.attributes}))

  def put_model_arrayset(self, mid, aid, input=True):
    """Starts a new model array set artifact, ready to receive data.
    
    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    aid: string, required
      Unique artifact identifier.
    input: boolean, optional
      Set to true (default) if array set is a model input.

    See Also
    --------
    :http:put:`/api/models/(mid)/arraysets/(aid)`
    """

    self.request("PUT", "/api/models/%s/arraysets/%s" % (mid, aid), 
      headers={"content-type":"application/json"}, data=json.dumps({"input":input}))

  def put_model_inputs(self, source, target):
    """Copies the input artifacts from one model to another. Both models 
    must be part of the same project. By default, array artifacts are copied 
    by reference instead of value for efficiency.

    Parameters
    ----------
    source: string, required
      Model ID source of artifacts.
    target: string, required
      Model ID target for artifacts.

    See Also
    --------
    :http:put:`/api/models/(mid)/inputs`
    """

    self.request("PUT", "/api/models/%s/inputs" % (target), 
      headers={"content-type":"application/json"}, data=json.dumps({"sid":source}))

  def put_model_parameter(self, mid, aid, value, input=True):
    """Store a model parameter artifact.

    Model parameters are JSON objects of arbitrary complexity.  They are stored directly within the model
    as part of its database record, so they should be limited in size (larger data should be stored using
    arraysets or files).

    To get the value of a parameter artifact, use :func:`get_model` and read the value
    directly from the model record.  An artifact id `foo` will be accessible in the
    record as `model["artifact:foo"]`.

    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    aid: string, required
      Unique (within the model) artifact id.
    value: object, required
      An arbitrary collection of JSON-compatible data.
    input: boolean, optional
      Marks whether this artifact is a model input.

    See Also
    --------
    :http:put:`/api/models/(mid)/parameters/(aid)`
    """

    self.request("PUT", "/api/models/%s/parameters/%s" % (mid, aid), 
      headers={"content-type":"application/json"}, data=json.dumps({"value":value, "input":input}))

  def put_project(self, pid, project):
    """Modifies a project.
    
    Parameters
    ----------
    pid: string, required
      Unique project identifier.
    project: dictionary, required
      JSON like dictionary with fields to modified model including:

        * name (*string, optional*)
        * description (*string, optional*)
        * acl (*string, optional*) -- access control list

    See Also
    --------
    :http:put:`/api/projects/(pid)`
    """
    return self.request("PUT", "/api/projects/%s" % pid, 
      headers={"content-type":"application/json"}, data=json.dumps(project))

  ##################################################################################
  # Convenience functions that layer additional functionality atop the RESTful API #
  ##################################################################################

  def find_project(self, name):
    """Return a project identified by name.

    Parameters
    ----------
    name: string, required
      The name of the project to return.

    Returns
    -------
    project: The matching project, which is an arbitrary collection of JSON-compatible data.

    Raises
    ------
    Exception
      If a project with a matching name can't be found, or more than one project matches the name.

    See Also
    --------
    :func:`find_or_create_project`, :func:`get_projects`
    """

    projects = [project for project in self.get_projects()["projects"] 
                if project["name"] == name]

    if len(projects) > 1:
      cherrypy.log.error("slycat.web.client.__init__.py find_project", 
        "More than one project matched the given name.")
      raise Exception("More than one project matched the given name.")
    elif len(projects) == 1:
      return projects[0]
    else:
      cherrypy.log.error("slycat.web.client.__init__.py find_project", 
        "No project matched the given name.")
      raise Exception("No project matched the given name.")

  def find_or_create_project(self, name, description=""):
    """Return a project identified by name, or newly created.

    Parameters
    ----------
    name: string, required
      The name of the project to return (or create).
    description: string, optional
      Description to use for the new project (if a new project is created).

    Returns
    -------
    pid: string
      Unique identifier of the matching (or newly created) project.

    Raises
    ------
    Exception
      If more than one project matches the given name.

    See Also
    --------
    :func:`post_projects`
    """

    projects = [project for project in self.get_projects()["projects"] 
                if project["name"] == name]
    
    if len(projects) > 1:
      cherrypy.log.error("slycat.web.client.__init__.py find_or_create_project", 
        "More than one project matched the given name. Try using a " +
        "different project name instead.")
      raise Exception("More than one project matched the given name. " +
        "Try using a different project name instead.")
    elif len(projects) == 1:
      return projects[0]["_id"]
    else:
      return self.post_projects(name, description)

  def update_model(self, mid, **kwargs):
    """Update model state.

    This function provides a more convenient alternative to :func:`put_model`.

    See Also
    --------
    :func:`put_model`
    """

    model = {key: value for key, value in list(kwargs.items()) if value is not None}
    self.put_model(mid, model)

  def join_model(self, mid):
    """Wait for a model to complete before returning.

    A Slycat model goes through several distinct phases over its lifetime:

    1. The model is created.
    2. Input artifacts are pushed into the model.
    3. The model is marked "finished".
    4. Optional one-time computation is performed on the server, storing output artifacts.
    5. The model is complete and ready to be viewed.

    Use this function in scripts that have performed steps 1, 2, and 3 and need to wait until
    step 4 completes.

    Parameters
    ----------
    mid: string, required
      Unique model identifier.

    Notes
    -----
    A model that hasn't been finished will never complete - you should
    ensure that post_model_finish() is called successfully before calling
    join_model().

    See Also
    --------
    :func:`post_model_finish`
    """

    while True:
      model = self.request("GET", "/api/models/%s" % (mid), 
                           headers={"accept":"application/json"})
      if "state" in model and model["state"] not in ["waiting", "running"]:
        return
      time.sleep(1.0)

  ##############################################################
  # Functions that manage file uploading using the RESTful API #
  ##############################################################

  def post_uploads(self, mid, parser, aids, input=True):
    """Create Slycat file upload session.

    Create an upload session used to upload files for storage as model artifacts. 
    Once an upload session has been created, use upload_file put_upload_file_part
    to upload files directly from the client to the server.

    Parameters
    ----------
    mid: string, required
      Unique model identifier.
    parser: string,
      Name of parser to call after completion of upload.
    aids: array, required
      Artifact IDs for storage.
    input: boolean, optional
      True to create input artifacts for the model.

    Returns
    -------
    uid: string
      Upload session ID.

    See Also
    --------
    :http:post:`/api/uploads`
    """

    return self.request("POST", "/api/uploads", headers={"content-type":"application/json"},
      data=json.dumps({"mid":mid, "input":input, "parser":parser, "aids":aids}))["id"]

  def put_upload_file_part(self, uid, pid, fid, file_slice):
    """Upload a file or part of a file to Slycat.

    Upload a file (or part of a file) as part of an upload session created with 
    post_uploads.

    Use the “pid” and “fid” parameters to specify that the data being uploaded is 
    for part M of file N. To upload a file from the client, specify the “file” parameter. 

    Parameters
    ----------
    uid: string, required
      Unique file upload session ID.
    pid: string, required
      Zero-based part ID of file being uploaded.
    fid: string, required
      Zero-based file ID for file being uploaded.
    file: file part in bytes, required
      File part to upload.

    See Also
    --------
    :http:put:`/api/uploads/(uid)/files/(fid)/parts/(pid)`
    """
    
    return self.request("PUT", "/api/uploads/%s/files/%s/parts/%s" % (uid, fid, pid),
      data={'file': base64.b64encode(file_slice)})

  def post_upload_finished(self, uid, file_parts):
    """Notify Slycat server that file upload is finished

    Notify the server that all files have been uploaded for the given upload 
    session, and processing can begin. The request must include the uploaded 
    parameter, which specifies the number of files that were uploaded, 
    and the number of parts in each file. The server uses this information to 
    validate that it received every part of every file that the client sent.

    Parameters
    ----------
    uid: string, required
      Unique file upload session ID.
    file_parts: array, required
      Array of length number of files, 
      with each entry number of slices for that file.

    See Also
    --------
    :http:post:`/api/uploads/(uid)/finished`
    """

    return self.request("POST", "/api/uploads/%s/finished" % uid, 
      headers={"content-type":"application/json"},
      data=json.dumps({'uploaded': file_parts}))

  def delete_upload(self, uid):
    """Delete uploaded files from Slycat server

    Delete an upload session used to upload files for storage as model artifacts. 
    This function must be called once the client no longer needs the session, 
    whether the upload(s) have been completed successfully or the client is cancelling 
    an incomplete session.

    Note that you can examine the return codes to see if the files have been completely
    parsed by Slycat.

    Parameters
    ----------
    uid: string, required
      Unique file upload session ID.

    Returns
    -------
    status_code: int
      Http status code:

        * 204 No Content – The upload session and temporary storage has been deleted.
        * 409 Conflict – The upload session cannot be deleted, parsing is in progress. 

    See Also
    --------
    :http:post:`/api/uploads`
    """
  
    # call directly to get response code
    response = self.session.delete(self.host + "/api/uploads/%s" % uid, 
                    proxies=self.keywords.get("proxies"), 
                    verify=self.keywords.get("verify"))
    
    # return code
    return response.status_code

  def upload_files(self, mid, file_list, parser, parser_parms, progress=True):
    """Upload a list of files for a given model to the Slycat server.
    The files will be parsed by specified parser.  This is not a direct call
    to the API, but rather a convenience call to load/parse a list of files.

    Parameters
    ----------
    mid: string, required
      Model ID associated with the files to be uploaded.
    file_list: string array, required
      Local files to be uploaded to the Slycat server.  This can be one file,
      but it should be passed as an array with one file.
    parser: string, required
      Name of parser to use on the Slycat server.
    parser_parms: array, required
      Any parameters to be passed to the parser (passed using aids when the 
      upload session is established).
    progress: boolean, optional
      Display a download progress indicator using standard out (will not
      be logged to the log file).
    """

    # how many files?
    num_files = len(file_list)

    # get number of parts for progress bar
    file_slices_to_upload = []
    for fid in range(num_files):
      num_slices = math.ceil(os.path.getsize(file_list[fid]) / self.file_slice_size)
      file_slices_to_upload.append(num_slices)

    # create upload session
    uid = self.post_uploads(mid, parser, parser_parms)

    # keep track of slices uploaded
    file_slices_uploaded = [0] * num_files

    # upload each file
    for fid in range(num_files):
      
      # print progress bar if desired
      if progress:
        print('Uploading "%s":' % file_list[fid])

      # split each file into slices
      with open(file_list[fid], "rb") as file:
        
        # get file slice
        file_slice = file.read(self.file_slice_size)
        while (file_slice != b''):

          # upload slice
          self.put_upload_file_part(uid, file_slices_uploaded[fid], fid, file_slice)

          # advance to next slice
          file_slice = file.read(self.file_slice_size)
          file_slices_uploaded[fid] += 1

          if progress:
            _print_progress_bar(file_slices_uploaded[fid], file_slices_to_upload[fid], 
              prefix = 'Progress:', suffix = 'Complete', length = 50)

    # finish upload
    self.post_upload_finished(uid, file_slices_uploaded)
      
    # wait for parsing to finish
    while True:
      if self.delete_upload(uid) == 409:
        time.sleep(1.0)
      else:
        break

def connect(arguments, **keywords):
  """Factory function for client connections that takes an option parser as input.
  
  Parameters
  ----------
  arguments: argument parser object, required
    Parsed command line arguments. 
  keywords: dictionary, optional
    Additional options to pass to requests.
  
  Returns
  -------
  connection: session for submitting requests to Slycat sever.
  """

  # arguments are from the command line parser,
  # keywords get passed into the actual requests

  # check for --no-verify or security certificate
  if arguments.no_verify:
    keywords["verify"] = False
  elif arguments.verify is not None:
    keywords["verify"] = arguments.verify

  return Connection(auth=(arguments.user, arguments.password),
    host=arguments.host, port=arguments.port, kerberos=arguments.kerberos,
    proxies={"http":arguments.http_proxy, "https":arguments.https_proxy}, 
    file_slice_size=arguments.file_slice_size, **keywords)


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# Creates dial-a-cluster run chart models by processing TDMS formatted 
# files and creating a dial-a-cluster model.  The model can be saved
# in the dac genric format and/or pushed directly to a Slycat server.

# S. Martin
# 8/30/2022

# connection to Slycat server
import slycat.web.client

# DAC upload
import slycat.web.client.dac_gen as dac_gen

# common tdms operations
import slycat.web.client.dac_tdms_util as dac_tdms_util
from slycat.web.client.dac_tdms_util import TDMSUploadError

# file name manipulation
import os

# manipulating command line arguments
from argparse import Namespace

# processing tables
import numpy as np
from natsort import natsorted
from copy import deepcopy

# compute PCA with sklearn
from sklearn.decomposition import PCA

# reading tdms files
import nptdms
import pandas as pd

# output zip file
from zipfile import ZipFile

# the group run chart variables allowed
RUN_CHART_MATCHES = ['TAD', 'Rp', 'Ip', 'DBV']
RUN_CHART_UNITS = ['nsec', 'kOhms', 'Amps', 'V']

# organize tdms files based on command line inputs
def catalog_tdms_files (arguments, log):

    # read all meta data before further run chart specific filtering
    metadata = dac_tdms_util.catalog_tdms_files(arguments, log)

    # get a list of tdms types and number of occurences
    tdms_types_occ = {}
    for row in range(len(metadata)):
        for row_type in metadata[row]["tdms_types"]:
            if row_type not in tdms_types_occ:
                tdms_types_occ[row_type] = 1
            else:
                tdms_types_occ[row_type] += 1

    # if --min_tdms_files is 0 then include only types occuring in all directories
    min_tdms_files = len(metadata)
    if arguments.min_tdms_files > 0:
        min_tdms_files = arguments.min_tdms_files
    
    # check that --min_tdms_files is not greater than number of directories
    if min_tdms_files > len(metadata):
        log("Warning: --min_tdms_files value exceeds number of directories, defaulting to 0.")
        min_tdms_files = len(metadata)

    # keep tdms types according to --min-tdms-files argument
    common_tdms_types = []
    intersection_flag = False
    for tdms_type in tdms_types_occ:
        if tdms_types_occ[tdms_type] >= min_tdms_files:
            common_tdms_types.append(tdms_type)
        else:
            intersection_flag = True

    # sort common types
    common_tdms_types = sorted(common_tdms_types)

    # remove files with uncommon types
    if intersection_flag:

        for row in range(len(metadata)):
            for tdms_ind in reversed(range(len(metadata[row]["tdms_types"]))):
                if metadata[row]["tdms_types"][tdms_ind] not in common_tdms_types:

                    # warn user we are ignoring this file
                    log('Ignoring file "' + metadata[row]["tdms_files"][tdms_ind] +
                        '" because type "' + metadata[row]["tdms_types"][tdms_ind] +
                        '" is not present in all runs.')

                    # remove files from catalog
                    metadata[row]["tdms_types"].pop(tdms_ind)
                    metadata[row]["tdms_files"].pop(tdms_ind)

    # sort catalog by tdms type
    for row in range(len(metadata)):

        # get tmds types and files
        row_tdms_types = metadata[row]["tdms_types"]
        row_tdms_files = metadata[row]["tdms_files"]

        # get indices of sorted tdms types
        tdms_inds = np.argsort(row_tdms_types)

        # sort and re-store types and files
        metadata[row]["tdms_types"] = list(np.asarray(row_tdms_types)[tdms_inds])
        metadata[row]["tdms_files"] = list(np.asarray(row_tdms_files)[tdms_inds])

    # tally suffixes used
    suffixes = set()
    for data in metadata:
        for suffix in data['include_suffixes']:
            suffixes.add(suffix)

    # sort suffixes
    suffixes = list(suffixes)
    suffixes.sort()

    # check any combined suffixes
    if arguments.combine:
        for i in range(len(arguments.combine)):
            combine_i = arguments.combine[i]

            # check that each combined run chart is at least a pair
            if len(combine_i) < 2:
                raise TDMSUploadError("--combine flag used with less than two TDMS suffixes, " + 
                    "please try again with at least two suffixes.")

            # check that combinations are in list of suffixes
            for j in range(len(combine_i)):
                if combine_i[j] not in suffixes:
                    raise TDMSUploadError('TDMS file suffix "' + combine_i[j] +
                        '" to combine was not found in the data directory.')
                
    # keep track of what we are including/excluding
    if arguments.exclude:
        log("Excluding TDMS file suffixes:")
        for suffix in arguments.exclude:
            log("\t%s" % suffix)

    log("Including TDMS file suffixes:")
    for suffix in suffixes:
        log("\t%s" % suffix)

    return metadata, common_tdms_types

# read all tdms run chart data
def read_tdms_files(metadata, run_chart_matches, common_tdms_types, log):

    # go through each row in the table
    skipped_rows = np.zeros(len(metadata))
    for row in range(len(metadata)):

        root_properties = []
        run_charts = []
        shot_numbers = []
        module_ID = []
        module_SN = []
        # module_ID_SN = []
        run_chart_headers = []

        for tdms_file_ind in range(len(metadata[row]["tdms_files"])):

            # report progress
            log('Reading .tdms file: "' + metadata[row]["tdms_files"][tdms_file_ind] + '".')

            # open tdms file
            tdms_file_name = metadata[row]["tdms_files"][tdms_file_ind]
            tdms_file_path = os.path.join(metadata[row]["source"], tdms_file_name)
            with nptdms.TdmsFile.open(tdms_file_path) as tdms_file:

                # save the root properties for the .tdms file
                root_properties.append(pd.DataFrame(tdms_file.properties, index=[0]))

                # get group properties
                group_header = None
                group_rows = []
                for group in tdms_file.groups():

                    # check that header is the same
                    if group_header is None:
                        group_header = list(group.properties.keys())

                    # this should never happen
                    elif group_header != list(group.properties.keys()):
                        raise TDMSUploadError('Group headers in file "' + tdms_file_name + 
                                        '" do not match.')

                    # keep track of table values
                    group_rows.append(list(group.properties.values()))

                # create pandas dataframe for group properies
                group_df = pd.DataFrame(data=group_rows, columns=group_header)

                # find numeric columns
                numeric_group_df = group_df.select_dtypes('float64')

                # select headers that have matching run charts
                header_matches = []
                rc_matches = []
                for header in numeric_group_df.head():
                    for run_chart in run_chart_matches:
                        if run_chart in header:
                            header_matches.append(header)
                            rc_matches.append(run_chart)

                # check that run charts have enough data
                if any(pd.notna(numeric_group_df[header_matches]).sum()==0):
                    log('Incomplete run chart data found for "' + tdms_file_path + '" -- ' +
                        'skipping directory "' + metadata[row]["source"] + '".')
                    skipped_rows[row] = 1

                # get indices of sorted tdms types
                rc_inds = np.argsort(rc_matches)

                # re-order run charts and headers alphabetically
                rc_matches = list(np.asarray(rc_matches)[rc_inds])
                header_matches = list(np.asarray(header_matches)[rc_inds])

                # check that run chart data is non-empty
                if rc_matches == []:
                    log('No run chart data found for "' + tdms_file_path + '" --' +
                        'skipping directory "' + metadata[row]["source"] +'".')
                    skipped_rows[row] = 1
                
                else:
                    # get sub dataframe with matching headers
                    run_charts.append(numeric_group_df[header_matches])

                    # also save run chart header matches
                    run_chart_headers.append(rc_matches)

                    # also save shot numbers
                    shot_numbers.append(group_df["Shot Number"])

                    # check for module SN
                    module_ID_i = check_module_col(group_df, "Module ID", tdms_file_path, log)
                    module_ID.append(module_ID_i)
                    module_SN_i = check_module_col(group_df, "Module SN", tdms_file_path, log)
                    module_SN.append(module_SN_i)
                    # module_ID_SN.append("/".join([str(module_ID_i), str(module_SN_i)]))

        # add to metadata
        metadata[row]["root_properties"] = root_properties
        metadata[row]["run_charts"] = run_charts
        metadata[row]["run_chart_headers"] = run_chart_headers
        metadata[row]["shot_numbers"] = shot_numbers
        metadata[row]["module_ID"] = module_ID
        metadata[row]["module_SN"] = module_SN
        # metadata[row]["module_ID_SN"] = module_ID_SN

    # remove skipped rows
    for row in reversed(range(len(metadata))):
        if skipped_rows[row]:
            metadata.pop(row)

    # check that data is still non-empty
    if metadata == []:
        raise TDMSUploadError("Too many run charts were missing.  " + 
            "No data remaining to be processed.")

    # go through metadata and identify run chart/root property headers
    # according to common tdms types
    common_run_chart_headers = [[] for i in range(len(common_tdms_types))]
    common_run_chart_roots = [[] for i in range(len(common_tdms_types))]
    common_root_prop_headers = [[] for i in range(len(common_tdms_types))]

    for tdms_type in range(len(common_tdms_types)):
        for row in range(len(metadata)):
            for col in range(len(metadata[row]["tdms_types"])):
                if metadata[row]["tdms_types"][col] == common_tdms_types[tdms_type]:

                    # check root properties headers
                    if not common_root_prop_headers[tdms_type]:
                        common_root_prop_headers[tdms_type] = \
                            list(metadata[row]["root_properties"][col].columns)
                    elif list(metadata[row]["root_properties"][col].columns) != \
                        common_root_prop_headers[tdms_type]:
                        raise TDMSUploadError('Found non-matching root property headers in directory"' +
                            metadata[row]["source"] + '", can not continue.')

                    # check run chart headers with units
                    if not common_run_chart_headers[tdms_type]:
                        common_run_chart_headers[tdms_type] = \
                            list(metadata[row]["run_charts"][col].columns)
                    elif list(metadata[row]["run_charts"][col].columns) != \
                        common_run_chart_headers[tdms_type]:
                        raise TDMSUploadError('Found non-matching run chart data in directory"' +
                            metadata[row]["source"] + '", can not continue.')

                    # check run chart headers no units
                    if not common_run_chart_roots[tdms_type]:
                        common_run_chart_roots[tdms_type] = \
                            metadata[row]["run_chart_headers"][col]
                    elif metadata[row]["run_chart_headers"][col] != \
                        common_run_chart_roots[tdms_type]:
                        raise TDMSUploadError('Found non-matching run chart data in directory"' +
                            metadata[row]["source"] + '", can not continue.')

    # go through metadata and add missing entries for each common tdms type
    for row in range(len(metadata)):
        for tdms_type in range(len(common_tdms_types)):
            if common_tdms_types[tdms_type] not in metadata[row]["tdms_types"]:

                # missing source tdms file
                metadata[row]["tdms_files"].insert(tdms_type, np.nan)

                # suffixes
                metadata[row]["include_suffixes"].insert(tdms_type, np.nan)

                # root properties
                nan_root_properties = pd.DataFrame(np.nan, index=[0], 
                    columns=common_root_prop_headers[tdms_type])
                metadata[row]["root_properties"].insert(tdms_type, nan_root_properties)

                # run charts
                nan_run_chart = pd.DataFrame(np.nan, index=[0], 
                    columns=common_run_chart_headers[tdms_type])
                metadata[row]["run_charts"].insert(tdms_type, nan_run_chart)

                # shot numbers
                nan_shot_numbers = pd.Series(np.nan, index=[0],
                    name="Shot Number")
                metadata[row]["shot_numbers"].insert(tdms_type, nan_shot_numbers)

                # module ID
                metadata[row]["module_ID"].insert(tdms_type, np.nan)

                # module SN
                metadata[row]["module_SN"].insert(tdms_type, np.nan)

        # use common tdms types for all data
        metadata[row]["tdms_types"] = common_tdms_types

        # use common run chart headers
        metadata[row]["run_chart_headers"] = common_run_chart_roots

    return metadata

# check for constant module ID/SN column
def check_module_col (group_df, header, tdms_file_path, log):

    # check for module ID
    if header not in group_df.columns:
        log(header + ' not found for "' + tdms_file_path + '" -- using NaN value.')
        module_ID_i = "NaN"

    # check that module ID is constant
    else:
        module_ID_i = group_df[header][0]
        for module_ID_j in group_df[header]:
            if module_ID_i != module_ID_j:
                log('Inconsistent ' + header + ' for "' + tdms_file_path + 
                    '" -- using NaN value.')
                module_ID_i = "NaN"

    return module_ID_i

# fill in metadata table
def read_metadata_table (metadata, arguments, log):

    # go through once to check that column names are the same
    root_head = list(metadata[0]["root_properties"][0].columns)
    for i in range(len(metadata)):
        root_props = metadata[i]["root_properties"]
        
        # check if root properties have same header
        for j in range(len(root_props)):
        
            root_props_j = list(root_props[j].columns)
            if root_props_j != root_head:
                
                # switch to longer root header
                if len(root_props_j) > len(root_head):
                    root_head = root_props_j
                    
    # remove shorter root headers?
    for i in reversed(range(len(metadata))):
        root_props = metadata[i]["root_properties"]

        # is it a short root header?
        for j in range(len(root_props)):
        
            root_props_j = list(root_props[j].columns)
            if root_props_j != root_head:
               
                # fill in missing columns with NaNs
                for k in range(len(root_head)):
                    if root_head[k] not in root_props_j:
                        root_props[j][root_head[k]] = np.nan

                log ('Found TDMS file with inconsistent root property header: "' +
                     metadata[i]["source"] + '" for "' + metadata[i]["tdms_types"][j] + 
                    '" -- using NaN values for missing columns.')

    # go through and get constant values
    constants = np.ones(len(root_head))
    for i in range(len(metadata)):
        root_props = metadata[i]["root_properties"]

        # check for constant values/nans
        for k in range(len(root_head)):
        
            const_val = root_props[0].iloc[0][root_head[k]]     
            for j in range(len(root_props)):
            
                # check for constant value
                curr_val = root_props[j].iloc[0][root_head[k]]
                if curr_val != const_val:
                    constants[k] = 0

    # collect constant/variable/nan columns
    const_cols = list(np.asarray(root_head)[np.where(constants)[0]])
    var_cols = list(np.asarray(root_head)[np.where(constants==0)[0]])

    # exclude "Configuration Data XML" column by default
    if not arguments.include_configuration_XML:

        # remove "Configuration DATA XML" from headers
        try:
            const_cols.remove("Configuration Data XML")
        except:
            pass
    
        try:
            var_cols.remove("Configuration Data XML")
        except:
            pass

    # construct metadata table header
    meta_column_names = ['Part', 'Lot', 'Batch', 'Serial Number', 'Source']
    for tdms_type in metadata[0]["tdms_types"]:
        meta_column_names += ['Module ID [' + tdms_type + ']']
        meta_column_names += ['Module SN [' + tdms_type + ']']

    meta_column_names += const_cols
    for tdms_type in metadata[0]["tdms_types"]:
        meta_column_names += [header + " [" + tdms_type + "]" for header in var_cols]

    # construct metadata table rows
    meta_rows = []
    for row in range(len(metadata)):

        # data from file names
        meta_row = [metadata[row]["part"], metadata[row]["lot"], metadata[row]["batch"],
            metadata[row]["sn"], metadata[row]["source"]]
        
        # module IDs/SNs
        meta_mod_interleaved = []
        for i in range(len(metadata[row]["module_ID"])):
            meta_mod_interleaved.append(metadata[row]["module_ID"][i])
            meta_mod_interleaved.append(metadata[row]["module_SN"][i])
        meta_row += meta_mod_interleaved

        # data from tdms files
        root_props = metadata[row]["root_properties"]

        # constant root properties        
        meta_row = meta_row + list(root_props[0].iloc[0][const_cols].values)

        # variable root properties
        for i in range(len(root_props)):
            meta_row = meta_row + list(root_props[i].iloc[0][var_cols].values)

        # add to table
        meta_rows.append(meta_row)

    # log results
    log("Collated metadata table with " + str(len(meta_column_names)) + " columns and " +
        str(len(meta_rows)) + " rows.")

    return meta_column_names, meta_rows

# construct units table
def read_units_table (metadata, arguments, log):

    # header is always the same for DAC units table
    meta_var_names = ['Name', 'Time Units', 'Units', 'Plot Type']

    # fill in variables.meta table
    meta_vars = []
    meta_units = dict()
    var_inds = dict()
    ind = 0
    for tdms_type in range(len(metadata[0]["tdms_types"])):
        for run_chart in metadata[0]["run_chart_headers"][tdms_type]:

            # name of run chart
            rc_name = run_chart + " [" + metadata[0]["tdms_types"][tdms_type] + "]"

            # units, default inferred
            units = "Not Given"
            if not arguments.do_not_infer_chart_units:
                units = RUN_CHART_UNITS[RUN_CHART_MATCHES.index(run_chart)]
            
            # construct row of table
            if arguments.curve:
                meta_vars.append([rc_name, 'Shot', units, 'Curve'])
            else:
                meta_vars.append([rc_name, 'Shot', units, 'Scatter'])

            # keep track of units for each tdms type
            meta_units[metadata[0]["tdms_types"][tdms_type]] = units

            # keep track of indices for variables
            var_inds[rc_name] = ind

            # next row
            ind = ind + 1

    # add any combined curves
    rc_inds = []
    if arguments.combine:

        # go through each combination and construct each run chart
        for i in range(len(arguments.combine)):

            # document combinations
            log("Combining TDMS types: " + str([j for j in arguments.combine[i]]))
            
            # get units
            combine_i = arguments.combine[i]
            units = meta_units[combine_i[0]]

            # check that units are the same for each TDMS type in combination
            combined_headers = []
            for j in range(len(combine_i)):
                if meta_units[combine_i[j]] != units:
                    raise TDMSUploadError('Inconsistent units found between combined TDMS types "' +
                        str(combine_i) + ".")
                combined_headers.append(combine_i[j])
            
            # truncate headers to first word or combine_length
            trunc_headers = []
            for j in range(len(combined_headers)):
                trunc_header = combined_headers[j].split()[0]
                trunc_header = trunc_header[0:arguments.combine_length]
                trunc_headers.append(trunc_header)
            combined_header = '/'.join(trunc_headers)

            # make a row for each run chart
            for run_chart in metadata[0]["run_chart_headers"][tdms_type]:
                rc_name = run_chart + " [" + combined_header + "]"
                if arguments.curve:
                    meta_vars.append([rc_name, 'Shot', units, 'Curve'])
                else:
                    meta_vars.append([rc_name, 'Shot', units, 'Scatter'])
                
                # keep track of indices into variables for combinations
                rc_inds_row = []
                for j in range(len(combined_headers)):
                    rc_inds_row.append(var_inds[run_chart + " [" + combined_headers[j] + "]"])
                rc_inds.append(rc_inds_row)

    # note if we inferred units
    if not arguments.do_not_infer_chart_units:
        log('Inferred run chart units from names.')

    return meta_var_names, meta_vars, rc_inds

# construct time step vectors
def read_timesteps (metadata):

    # go through data and find maximum length of each run chart
    time_steps_max = []
    for row in range(len(metadata)):

        # gather time steps per row
        time_steps_row = []
        for tdms_type in range(len(metadata[0]["tdms_types"])):
            time_steps_row = time_steps_row + \
                [metadata[row]["run_charts"][tdms_type].shape[0]] * \
                len(metadata[0]["run_chart_headers"][tdms_type])
        
        # get maximum time steps for all rows
        if row == 0:
            time_steps_max = time_steps_row
        else:
            time_steps_max = np.maximum(time_steps_max, time_steps_row)

    # construct time steps matrix
    time_steps = []
    for max in time_steps_max:
        time_steps.append(list(np.asarray(range(max)) + 1))

    return time_steps

# construct shot numbers vectors instead of time steps
def read_shot_numbers (metadata):

    # go through data and construct shot number unions
    shot_numbers = []
    for row in range(len(metadata)):

        # combine shot numbers (union)
        if row == 0:
            shot_numbers = [set(metadata[row]["shot_numbers"][tdms_type]) for
            tdms_type in range(len(metadata[0]['tdms_types']))]
        else:
            shot_numbers = [shot_numbers[tdms_type].union(
                set(metadata[row]["shot_numbers"][tdms_type])) for
                tdms_type in range(len(metadata[0]['tdms_types']))]

    # remove nan values 
    for row in range(len(shot_numbers)):
        no_nan_shot_numbers = {x for x in shot_numbers[row] if pd.notna(x)}
        shot_numbers[row] = no_nan_shot_numbers

    # sort resulting shot numbers using natural sort
    shot_numbers = [[natsorted(list(shot_numbers[tdms_type]))] * 
        len(metadata[0]["run_chart_headers"][tdms_type]) for 
        tdms_type in range(len(metadata[0]['tdms_types']))]

    return shot_numbers

# construct variable matrices
def read_variable_matrices (metadata, time_steps, arguments, log):

    # put maximum number of time steps
    # in format of run chart data
    max_time_steps = []
    k = 0
    for i in range(len(metadata[0]["tdms_types"])):
        run_chart_max_time_steps = []
        for j in range(len(metadata[0]["run_chart_headers"][i])):
            run_chart_max_time_steps.append(len(time_steps[k]))
            k = k + 1
        max_time_steps.append(run_chart_max_time_steps)

    # variable matrices are a list of matrices, one for each variable
    # rows are variable values, columns are time
    var_data = []
    var_data_nan = []
    inferred_variables = False
    use_const_value = False
    for tdms_type in range(len(metadata[0]["tdms_types"])):
        for run_chart in range(len(metadata[0]["run_chart_headers"][tdms_type])):
            variable_i = []
            variable_i_nan = []
            for row in range(len(metadata)):
                
                # get run chart data
                run_chart_data = list(metadata[row]["run_charts"][tdms_type].iloc[:,run_chart])
                run_chart_data_nan = list(metadata[row]["run_charts"][tdms_type].iloc[:,run_chart])

                # check if there are any actual numbers in the run chart
                #if len(run_chart_data) > 1 or not np.isnan(run_chart_data[0]):
                # for identically nan vectors, use_const_value
                if np.isnan(run_chart_data).all():
                    if arguments.use_const_value:

                        # fill out run chart data with constant value
                        run_chart_data = np.empty ((max_time_steps[tdms_type][run_chart]))
                        run_chart_data[:] = arguments.use_const_value[0]

                        # fill out run chart data with NaNs
                        run_chart_data_nan = np.empty ((max_time_steps[tdms_type][run_chart]))
                        run_chart_data_nan[:] = np.nan

                        use_const_value = True

                    else:
                        raise TDMSUploadError('Found absent run chart in source "' +
                            metadata[row]["source"] + '" for type "' + 
                            metadata[row]["tdms_types"][tdms_type] + '".  Use const value option' +
                            ' (e.g. --use-const-value) to specify value for absent run charts.')

                # check for NaN values
                elif np.isnan(run_chart_data).any():
                    log('Skipping NaN values in file "' + metadata[row]["tdms_files"][tdms_type] +
                        '" run chart "' + metadata[row]["run_charts"][tdms_type].columns[run_chart] + '".')
                    
                    # skip NaN values (we already checked that there 
                    # are some non-nan values when first reading tdms files)
                    run_chart_data = [x for x in run_chart_data if not np.isnan(x)]
                    run_chart_data_nan = [x for x in run_chart_data_nan if not np.isnan(x)]

                # extend run chart data by last value
                if len(run_chart_data) < max_time_steps[tdms_type][run_chart]:
                    if arguments.infer_last_value:
                        run_chart_data = run_chart_data + [run_chart_data[-1]] * \
                            (max_time_steps[tdms_type][run_chart] - len(run_chart_data))
                        run_chart_data_nan = run_chart_data_nan + [np.nan] * \
                            (max_time_steps[tdms_type][run_chart] - len(run_chart_data_nan))
                        inferred_variables = True
                    else:
                        raise TDMSUploadError('Found run chart length discrepancy in file "' +
                            metadata[row]["tdms_files"][tdms_type] + '".  Use inference options' +
                            ' (e.g. --infer-last-value) to correct lengths by inferring data.')
    
                # add run chart to current variable
                variable_i.append(run_chart_data)
                variable_i_nan.append(run_chart_data_nan)

            # add variable matrix to list
            var_data.append(variable_i)
            var_data_nan.append(variable_i_nan)

    if inferred_variables:
        log("Variables inferred to correct run chart length discrepancies.")

    if use_const_value:
        log('Used constant value "' + str(arguments.use_const_value[0]) + '" for missing variables.')

    return var_data, var_data_nan

# get variables for shot numbers
def read_variable_shot_matrices (metadata, shot_numbers, combined_inds, arguments, log):

    # put maximum number of shots
    # in format of run chart data
    max_shot_numbers = []
    for i in range(len(shot_numbers)):
        run_chart_max_shot_numbers = []
        for j in range(len(shot_numbers[i])):
            run_chart_max_shot_numbers.append(len(shot_numbers[i][j]))
        max_shot_numbers.append(run_chart_max_shot_numbers)

    # variable matrices are number variables by number files by number shots
    # when using shot numbers var data will almost certainly contain NaNs
    # however, we still need inferred variables for PCA
    var_data = []
    var_data_nan = []
    inferred_variables = False
    use_const_value = False
    for tdms_type in range(len(metadata[0]["tdms_types"])):
        for run_chart in range(len(metadata[0]["run_chart_headers"][tdms_type])):

            variable_i = []
            variable_i_nan = []
            for row in range(len(metadata)):

                # get run chart data
                run_chart_data = list(metadata[row]["run_charts"][tdms_type].iloc[:,run_chart])
                run_chart_shots = metadata[row]["shot_numbers"][tdms_type]

                # fill out run chart data with NaNs
                run_chart_data_nan = np.empty ((max_shot_numbers[tdms_type][run_chart]))
                run_chart_data_nan[:] = np.nan

                # check if there are any actual numbers in the run chart
                if len(run_chart_shots) > 1 or not np.isnan(run_chart_shots[0]):

                    # replace known entries with numbers
                    for i in range(len(run_chart_shots)):
                        run_chart_data_nan[shot_numbers[tdms_type][run_chart].index(
                            run_chart_shots[i])] = run_chart_data[i]

                # infer variables for NaN values
                run_chart_data = np.copy(run_chart_data_nan)
                mask = np.isnan(run_chart_data)

                # for identically nan vectors, use_const_value
                if all(mask):
                    if arguments.use_const_value:
                        run_chart_data[mask] = arguments.use_const_value[0]
                        use_const_value = True
                    else:
                        raise TDMSUploadError('Found absent run chart in source "' +
                            metadata[row]["source"] + '" for type "' + 
                            metadata[row]["tdms_types"][tdms_type] + '".  Use const value option' +
                            ' (e.g. --use-const-value) to specify value for absent run charts.')

                elif any(mask):
                    if arguments.infer_last_value:
                        run_chart_data[mask] = np.interp(np.flatnonzero(mask), 
                            np.flatnonzero(~mask), run_chart_data[~mask])
                        inferred_variables = True
                    else:
                        raise TDMSUploadError('Found run chart length discrepancy in file "' +
                            metadata[row]["tdms_files"][tdms_type] + '".  Use inference options' +
                            ' (e.g. --infer-last-value) to correct lengths by inferring data.')

                # add run chart to current variable
                variable_i.append(run_chart_data.tolist())
                variable_i_nan.append(run_chart_data_nan.tolist())

            # add variable matrix to list
            var_data.append(variable_i)
            var_data_nan.append(variable_i_nan)

    # we also have to convert time steps to numbers (e.g. strip a/b/c values)
    if not arguments.highlight_shot_numbers:
        shot_numbers = strip_abc(shot_numbers)
    
    # otherwise we convert as nan values 
    else:
        shot_numbers = convert_abc(shot_numbers)

    time_steps = []
    for i in range(len(shot_numbers)):
        time_steps = time_steps + shot_numbers[i]

    if inferred_variables:
        log("Variables inferred to correct run chart shot number discrepancies.")
    
    if use_const_value:
        log('Used constant value "' + str(arguments.use_const_value[0]) + '" for missing variables.')

    # add combined variables, if requested
    if arguments.combine:

        # go through each combination to extend matrices
        for i in range(len(combined_inds)):
            combined_time_steps = []
            combine_i = combined_inds[i]

            # check that time steps are non-intersecting
            for j in range(len(combine_i)):
                if time_steps[combine_i[j]] not in combined_time_steps:
                    combined_time_steps += time_steps[combine_i[j]]
                else:
                    raise TDMSUploadError('Shot number intersection for combined TDMS types, ' +
                        'please try a different combination.')

            # check that time steps are in order
            if sorted(combined_time_steps) != combined_time_steps:
                raise TDMSUploadError('Shot numbers are out of order for combined TDMS types, ' +
                    'please try a different combination order.')

            # add combined_time_steps to time steps matrix
            time_steps.append(combined_time_steps)

            # add variables to variable matrices
            combined_var_data = deepcopy(var_data[combine_i[0]])
            combined_var_data_nan = deepcopy(var_data_nan[combine_i[0]])
            for j in range(1, len(combine_i)):
                for k in range(len(combined_var_data)):
                    combined_var_data[k] = combined_var_data[k] + var_data[combine_i[j]][k]
                    combined_var_data_nan[k] = combined_var_data_nan[k] + var_data_nan[combine_i[j]][k]
                                                                                       
            # append combined matrix to variable matrices
            var_data.append(combined_var_data)
            var_data_nan.append(combined_var_data_nan)

    return var_data, var_data_nan, time_steps

# helper function for read_variable_shot_matrices
# converts strings to integers and converts a/b/c to fractions
# for example '31' would be 31 and '31a' would be 31.25
def strip_abc(item):

    if isinstance(item, list):
        return [strip_abc(x) for x in item]
    
    else:
        if item[-1] in ['a', 'b', 'c']:

            # convert 'a' to 1/4, 'b' to 1/2, 'c' to 3/4
            suffix = (['a', 'b', 'c'].index(item[-1]) + 1)/4

            try:
                return int(item[:-1]) + suffix
            except:
                raise ValueError ("Shot number not an integer.")
        else:

            try:
                return int(item)
            except:
                raise ValueError ("Shot number not an integer.")

# helper function for read_variable_shot_matrices
# converts strings to integers and converts a/b/c to nan
def convert_abc(item):

    if isinstance(item, list):
        return [convert_abc(x) for x in item]
    
    else:
        if item[-1] in ['a', 'b', 'c']:
            return np.nan
        
        else:

            try:
                return int(item)
            except:
                raise ValueError ("Shot number not an integer.")

# compute PCA for variables
def compute_PCA (var_data, arguments, log):

    # create pairwise distance matrix, using PCA or landmarks if available
    var_dist = []
    for i in range(len(var_data)):

        # compute PCA using sklearn
        pca = PCA(n_components=arguments.num_PCA_comps)

        try:
            dist_i = pca.fit_transform(var_data[i])

        except ValueError:
            dist_i = np.zeros ((len(var_data[i]), arguments.num_PCA_comps))
            log('Warning: could not perform PCA, too few components -- using zero values.')

        # save PCA projections
        var_dist.append(dist_i)

    return var_dist

# write data to given output directory
def write_dac_gen(meta_col_names, meta_rows, meta_var_names, meta_vars, 
            time_steps, var_data, var_dist, arguments):

    # create output directory if it doesn't already exist
    if os.path.exists(arguments.output_zip_file):
        log('Warning: overwriting .zip file "' + arguments.output_zip_file + '".')

    # write out zip file
    with ZipFile(arguments.output_zip_file, "w") as zip_file:

        # convert metadata table to string
        metadata_table = ','.join(meta_col_names) + "\n" + mat2str(meta_rows)

        # write out metadata table as "datapoints.dac"
        zip_file.writestr("datapoints.dac", metadata_table)

        # convert units table to string
        units_table = ','.join(meta_var_names) + '\n' + mat2str(meta_vars)

        # write out variables units table
        zip_file.writestr(os.path.join("var","variables.meta"), units_table)
        
        # write out pca.csv file with number of components
        zip_file.writestr("pca.csv", str(arguments.num_PCA_comps))
        
        # write out times/variables/distances
        num_vars = len(time_steps)
        for i in range(num_vars):

            # convert time steps to string
            time_i = time_steps[i]
            time_i = ','.join([str(time_i[j]) for j in range(len(time_i))])

            # write out to zip file as time/variable_i
            zip_file.writestr(os.path.join("time","variable_" + str(i+1) + ".time"), time_i)

            # write out variable data to zip file
            zip_file.writestr(os.path.join("var","variable_" + str(i+1) + ".var"), 
                mat2str(var_data[i]))

            # write out variable distances to zip file
            zip_file.writestr(os.path.join("dist","variable_" + str(i+1) + ".dist"), 
                mat2str(var_dist[i]))

    log('Output written to .zip file "' + arguments.output_zip_file + '".')

# helper function to convert a matrix to a comma separatedstring
def mat2str (mat):

    return '\n'.join([','.join([str(mat[j][k]).replace('\n', ' ').replace('\r', '').strip()
        for k in range(len(mat[j]))])
        for j in range(len(mat))])

# check arguments and create model
def create_model(arguments, log):

    # if you have --highlight-shot-numbers, you also have to have --use-shot-numbers
    if arguments.highlight_shot_numbers:
        if not arguments.use_shot_numbers:
            raise TDMSUploadError('Must use "--use-shot-numbers" with "--highlight-shot-numbers".')
    
    # if you have --combine, you must also have --use-shot-numbers
    if arguments.combine:
        if not arguments.use_shot_numbers:
            raise TDMSUploadError('Must use "--use-shot-numbers" with "--combine".')

    # can't have both overvoltage and sprytron
    if arguments.overvoltage and arguments.sprytron:
        raise TDMSUploadError("Can't use both overvoltage and sprytron options " +
              "together. Please select one or the other and try again.")

    # check if zip file is correctly
    if not arguments.output_zip_file.endswith(".zip"):
        raise TDMSUploadError ('Must use .zip extension on output file name.')

    # organize tdms files
    metadata, common_tdms_types = catalog_tdms_files(arguments, log)

    # screen columns according to user input
    run_chart_matches = RUN_CHART_MATCHES
    if arguments.use_run_charts is not None:
        run_chart_matches = arguments.use_run_charts

    # report on run charts being used
    log("Using run charts in: " + str(run_chart_matches))

    # read tdms data
    metadata = read_tdms_files(metadata, run_chart_matches, common_tdms_types, log)

    # construct metadata table
    meta_col_names, meta_rows = read_metadata_table(metadata, arguments, log)

    # construct units table
    meta_var_names, meta_vars, combined_inds = read_units_table(metadata, arguments, log)

    # construct time/variable data
    if arguments.use_shot_numbers:
        shot_numbers = read_shot_numbers(metadata)
        var_data, var_data_nan, time_steps = \
            read_variable_shot_matrices (metadata, shot_numbers, combined_inds, arguments, log)
    else:
        time_steps = read_timesteps(metadata)
        var_data, var_data_nan = read_variable_matrices (metadata, time_steps, arguments, log)

    # compute PCA representation
    var_dist = compute_PCA (var_data, arguments, log)
    
    # save files, use nans in plot if requested
    write_dac_gen(meta_col_names, meta_rows, meta_var_names, meta_vars, 
                  time_steps, var_data, var_dist, arguments)

    # add output file for dac_gen script
    dac_gen_args = Namespace(**vars(arguments), **{'dac_gen_zip': arguments.output_zip_file})

    # push model using dac_gen
    if not arguments.do_not_upload:
        mid = dac_gen.upload_model(dac_gen_args, log)

        # supply the user with a direct link to the new model.
        host = arguments.host
        if arguments.port:
            host = host + ":" + arguments.port
        log("Your new model is located at %s/models/%s" % (host, mid))
        log('***** DAC Model Successfully Created *****')

    # should we erase the .zip file created
    if arguments.clean_up_output:
        os.remove(arguments.output_zip_file)
        log('Deleted file "' + arguments.output_zip_file + '".')

# logging is just printing to the screen
def log (msg):
    print(msg)

# set up argument parser
def parser ():

    # provide additional command line arguments for TDMS files
    parser = slycat.web.client.ArgumentParser(description=
        "Creates a Dial-A-Cluster model using run charts from .tdms files.")

    # input consists of tdms batch format or tdms file
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument("--input-tdms-glob", nargs=2,
        help="Use tdms directory glob format as input, first argument is directory, organized by " +
             "part and lot as on the production server; second argument is part_num_match, speficied with " +
             'a unix glob using wildcards, for example "XXXXXX*", or "XXXXXX_XX*". ' + 
             'Note you should use "" in Unix to pass wildcards. Produces one run chart per directory.')

    # input tdms batches
    group.add_argument("--input-tdms-batches", nargs=3, 
        help="Use tdms batch format as input, first argument is directory, organized by part and lot " +
             'as on the production server; second argument is part_lot, e.g. "XXXXXX_XX", ' +
             'where part and lot numbers should be constant; and third argument is batches, ' +
             'e.g. "1,3,4-6,11-24".  Use "*" to designate all batches. Produces ' + 
             "one run chart per directory.")

    # input tdms directory
    group.add_argument("--input-tdms-dir", nargs=1,
        help="Use all .tdms files in root directory as input. Produces one run chart per " +
             "subdirectory found containing .tdms files.")

    # unstructured option for tdms dir
    parser.add_argument("--unstructured", action="store_true",
        help="For unstructured directory names, infer run chart groups from file names instead.")
    
    # output file for dac generic format, mandatory
    parser.add_argument("output_zip_file",
        help="Output file name for writing dac generic format of the run " +
             " chart model.  If it already exists it will be overwritten.")

    # delete output file after successful model creation
    parser.add_argument("--clean-up-output", action="store_true",
        help="Delete output .zip file after successful model creation.")

    # do not upload to slycat
    parser.add_argument("--do-not-upload", action="store_true",
        help="Do not upload to slycat server.")

    # model and project names/descriptions
    parser.add_argument("--marking", default="cui", 
        help="Marking type.  Default: %(default)s")
    parser.add_argument("--model-description", default="", 
        help="New model description.  Default: %(default)s")
    parser.add_argument("--model-name", default="DAC Run Chart Model", 
        help="New model name.  Default: %(default)s")
    parser.add_argument("--project-description", default="", 
        help="New project description.  Default: %(default)s")
    parser.add_argument("--project-name", default="DAC Run Chart Models", 
        help="New project name.  Default: %(default)s")

    # spyrton/overvoltage parameters
    parser.add_argument("--overvoltage", action="store_true",
        help="Expecting overvoltage data.")
    parser.add_argument("--sprytron", action="store_true",
        help="Expecting sprytron data.")

    # run-chart file selection
    parser.add_argument("--use-run-charts", nargs="+", choices=RUN_CHART_MATCHES,
        help="Use specified variables for run-charts, otherwise run-charts " + 
             "are inferred from non-constant values.")

    # inferring data
    parser.add_argument("--do-not-infer-chart-units", action="store_true",
        help="Do not infer run chart units. Default is to infer units " +
             "from variable name.")
    parser.add_argument("--infer-last-value", action="store_true", 
        help="Fill out short run charts with last given value.")
    parser.add_argument("--use-const-value", nargs=1, type=float,
        help="Constant value to use for absent run charts (no default).")
    
    # exclude XML by default
    parser.add_argument("--include_configuration_XML", action="store_true",
        help='Include "Configuration Data XML" if present in data.')

    # compute using PCA model
    parser.add_argument("--num-PCA-comps", default=10, type=int,
        help="Number of PCA components to use, integer >= 2.  " +
             "Default: %(default)s")

    # exclude .tdms files
    parser.add_argument("--exclude", nargs="+",
        help='TDMS file suffixes to exclude. ' + 
             'If you want suffixes that include spaces, use quotes, ' +
             'e.g. "suffix with space".')

    # use curve plots for run charts
    parser.add_argument("--curve", action="store_true",
        help="Use curves for run charts instead of scatter plots.")

    # use shot numbers to line up plots
    parser.add_argument('--use-shot-numbers', action="store_true",
        help="Use actual shot numbers (default is to use shot number order only).")
    
    # add vertical lines to indicate non-numeric time steps
    parser.add_argument('--highlight-shot-numbers', action="store_true",
        help="Highlight non-numeric a/b/c shot numbers using vertical lines, otherwise they "+
             "are converted to values of (.25, .5, .75).")
    
    # combination time series
    parser.add_argument('--combine', nargs="+", action="append",
        help="Concatenate TDMS suffixes into a single run chart, e.g. " +
             '--combine "Acceptance Trigger Operation" "Factory Trigger Operation" ' +
             "would concatenate the run charts from those two TDMS shots into " +
             "a single run chart.  Can be use multiple times with different TDMS files.")
    parser.add_argument('--combine-length', default=10, type=int,
        help="Truncation length for each header in combined header.  Truncates to first " +
              'word or "--combine-length", whichever is less).')
    
    # minimum number of tdms types needed to keep variables
    parser.add_argument('--min-tdms-files', type=int, default=0, 
        help="Minimum number of tdms files necessary to include in the run chart.  Defaults " +
              "to 0, meaning all files must be present.  If files are not present, " +
              "the respective run charts are given 0 values for the dimension reduction " +
              "calculation and NaN values for the plots.")
    
    return parser

# tdms entry point
def main():

    # set up argument parser
    run_chart_parser = parser()  

    # get user arguments
    arguments = run_chart_parser.parse_args()

    # check arguments and create model
    create_model(arguments, log)

# command line version to load a single DAC model
if __name__ == "__main__":

    main()


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

# This script read and processes a .csv file before pushing it to the Slycat server
# as a parameter space model.

import os
import re

import numpy as np

import slycat.web.client
import urllib.parse as urlparse

# slycat csv parser
import slycat.pandas_util

# parse command line arguments
def parser():

  parser = slycat.web.client.ArgumentParser(description=
    "Create Slycat Parameter Space model from .csv table.")

  # input file
  parser.add_argument("file", help="Input CSV file.")

  # input/output columns
  parser.add_argument("--input-columns", default=[], nargs="*", 
    help="Input column names, can't be media columns.")
  parser.add_argument("--output-columns", default=[], nargs="*", 
    help="Output column names, can't be media columns.")

  # media columns
  parser.add_argument("--media-columns", default=None, nargs="*", 
    help="Columns that contain media URLs.")
  parser.add_argument("--media-hostname", default=None, 
    help="Optionally override the hostname for media columns.")
  parser.add_argument("--strip", type=int, default=None, 
    help="Optionally strip N prefix directories from the stored URLs.")

  # model and project names/descriptions
  parser.add_argument("--marking", default="cui", 
      help="Marking type.  Default: %(default)s")
  parser.add_argument("--model-description", default="", 
      help="New model description.  Default: %(default)s")
  parser.add_argument("--model-name", default="PS Model", 
      help="New model name.  Default: %(default)s")
  parser.add_argument("--project-description", default="", 
      help="New project description.  Default: %(default)s")
  parser.add_argument("--project-name", default="PS Models", 
      help="New project name.  Default: %(default)s")

  return parser

# logging is just printing to the screen
def log (msg):
    print(msg)

# create PS model, show progress by default, arguments 
# include the command line connection parameters
# and project/model information
def upload_model (arguments, attributes, dimensions, data, progress=True):

  # get column names/types
  column_names = [attribute["name"] for attribute in attributes]
  column_types = [attribute["type"] for attribute in attributes]

  # setup a connection to the Slycat Web Server.
  connection = slycat.web.client.connect(arguments)

  # create a new project to contain our model.
  pid = connection.find_or_create_project(arguments.project_name, arguments.project_description)

  # create the new, empty model.
  mid = connection.post_project_models(pid, "parameter-image", 
    arguments.model_name, arguments.marking, arguments.model_description)

  # Upload our observations as "data-table".
  connection.put_model_arrayset(mid, "data-table")

  # Start our single "data-table" array.
  connection.put_model_arrayset_array(mid, "data-table", 0, dimensions, attributes)

  # Upload each column into the array.
  for index in range(len(column_names)):
    values = [np.asarray(data[index])]

    # push to server
    connection.put_model_arrayset_data(mid, "data-table", "0/%s/..." % index, values)

  # Store the remaining parameters.
  input_col_inds = [index for index, header in enumerate(column_names) 
    if header in arguments.input_columns and column_types[index] != "string"]
  connection.put_model_parameter(mid, "input-columns", input_col_inds)
  output_col_inds = [index for index, header in enumerate(column_names) 
    if header in arguments.output_columns and column_types[index] != "string"]
  connection.put_model_parameter(mid, "output-columns", output_col_inds)
  media_col_inds = [index for index, header in enumerate(column_names) 
    if header in arguments.media_columns and column_types[index] == "string"]
  connection.put_model_parameter(mid, "image-columns", media_col_inds)

  # Signal that we're done uploading data to the model.  This lets Slycat Web
  # Server know that it can start computation.
  connection.post_model_finish(mid)

  # Wait until the model is ready.
  connection.join_model(mid)

  return mid

# check arguments and create model, returns URL to model
def create_model (arguments, log):

  # check that input file exists
  if not os.path.isfile(arguments.file):
    log("Input file " + arguments.file + " does not exist.")
    raise ValueError("Input file " + arguments.file + " does not exist.")

  # check that input/output columns are non-intersecting
  if set(arguments.input_columns).intersection(set(arguments.output_columns)):
    log("Input columns and output columns are intersecting.")
    raise ValueError("Input columns and output columns are intersecting.")

  # parse file using standard slycat csv parser
  attributes, dimensions, data, csv_read_error = \
    slycat.pandas_util.parse_file(arguments.file, file_name=True)

  # check for warnings/errors
  for i in range(len(csv_read_error)):
    if csv_read_error[i]["type"] == "warning":
      log("Warning: " + csv_read_error[i]["message"])
    else:
      log("Error: " + csv_read_error[i]["message"])
      raise ValueError(csv_read_error[i]["message"])

  # get column names/types
  column_names = [attribute["name"] for attribute in attributes]
  column_types = [attribute["type"] for attribute in attributes]

  # check that inputs are in headers
  if arguments.input_columns is not None:
    for input_col in arguments.input_columns:
      if input_col not in column_names:
        log('Input column "' + input_col + '" not found in input file.')
        raise ValueError('Input column "' + input_col + '" not found in input file.')

  # check that outputs are in headers
  if arguments.output_columns is not None:
    for output_col in arguments.output_columns:
      if output_col not in column_names:
        log('Output column "' + output_col + '" not found in input file.')
        raise ValueError('Output column "' + output_col + '" not found in input file.')
      
  # check that meda-columns are in headers
  if arguments.media_columns is not None:
    for media_col in arguments.media_columns:
      if media_col not in column_names:
        log('Media column "' + media_col + '" not found in input file.')
        raise ValueError('Media column "' + media_col + '" not found in input file.')

  # check that there is at least one numeric columns (to display scatter plot)
  if not 'float64' in column_types:
    log("You must supply at least one numeric column in the input data.")
    raise ValueError("You must supply at least one numeric column in the input data.")

  # identify media columns automatically if not provided
  if arguments.media_columns is None:
    arguments.media_columns = []

    # search for "file://"
    expression = re.compile("file://")
    search = np.vectorize(lambda x:bool(expression.search(x)))

    # go through each column
    for header in range(len(column_names)):

      # strings are stored in pandas object type
      if column_types[header] == "string":
        if np.any(search(data[header])):
          arguments.media_columns.append(column_names[header])

  # replace media URL hostnames, if requested
  if arguments.media_hostname is not None:
    
    # go through each column
    for header in range(len(column_names)):

      # if media column is given and a string
      if column_names[header] in arguments.media_columns and \
         column_types[header] == "string":
        modified = []

        # replace file://cluster with file://hostname
        for uri in data[header]:
          uri = urlparse.urlparse(uri)
          path = uri.path
          uri = "file://%s%s" % (arguments.media_hostname, path)
          modified.append(uri)
        data[header] = modified

  # strip prefix directories, if requested
  if arguments.strip is not None:

    # go through each column:
    for header in range(len(column_names)):

      # for media columns, strip prefix directories
      if column_names[header] in arguments.media_columns and \
         column_types[header] == "string":
        modified = []

        # strip prefixes
        for uri in df[header].values:
          uri = urlparse.urlparse(uri)
          hostname = uri.netloc
          path = os.path.join(*([os.path.abspath(os.path.dirname(arguments.file))] + 
            uri.path.split(os.sep)[arguments.strip + 1:]))
          uri = "file://%s%s" % (hostname, path)
          modified.append(uri)
        data[header] = modified

  # echo inputs to user
  log('Input file: ' + arguments.file)
  if arguments.input_columns is not None:
    log('Input columns: ' + str(arguments.input_columns))
  if arguments.output_columns is not None:
    log('Output columns: ' + str(arguments.output_columns))
  if arguments.media_columns != []:
    log('Media columns: ' + str(arguments.media_columns))
  if arguments.media_hostname is not None:
    log('Media hostname: ' + str(arguments.media_hostname))
  if arguments.strip is not None:
    log('Stripped ' + str(arguments.strip) + ' prefixes from media columns.')

  # upload model
  mid = upload_model (arguments, attributes, dimensions, data, progress=True)

  # supply direct link to model
  host = arguments.host
  if arguments.port:
      host = host + ":" + arguments.port
  url = host + "/models/" + mid
  log("Your new model is located at %s" % url)
  log('***** PS Model Successfully Created *****')

  # return url
  return url

# main entry point
def main():

    # set up argument parser
    ps_parser = parser()  

    # get user arguments
    arguments = ps_parser.parse_args()

    # check arguments and create model
    url = create_model(arguments, log)

# command line version
if __name__ == "__main__":

    main()


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# Creates dial-a-cluster models by uploading TDMS and TMDS .zip formatted 
# files to the Slycat server.  Once uploaded the models are created on the
# server using the TDMS parser.
#
# This code can be run independently or used as a module to batch processs
# TDMS files.
#
# S. Martin
# 10/21/2020

# connection to Slycat server
import slycat.web.client

# DAC UI defaults
from slycat.web.client.dac_defaults import dac_model_defaults
import slycat.web.client.dac_tdms_util as dac_tdms_util
from slycat.web.client.dac_tdms_util import TDMSUploadError

# file name manipulation
import os

# get suffixes from zip file
import zipfile

# check for .TDM, .tdms, or a single .zip file extension
# returns "not-tdms" for wrong extension, "tdms" for correct extensions,
# "zip" for a single .zip file
def check_file_extensions (file_list):

    # a single file can end with .zip, .TDM, or .tdms
    if len(file_list) == 1:

        # check for .zip file
        if file_list[0].lower().endswith(".zip"):
            return "zip"
        
        # check for .tdm or .tdms file
        if file_list[0].lower().endswith('.tdm') or \
           file_list[0].lower().endswith('.tdms'):
            return "tdms"

    # multiple files can only end with .TDM or .tdms
    ext_OK = "tdms"

    for file in file_list:
        if not file.lower().endswith('.tdm') and \
           not file.lower().endswith('.tdms'):
            ext_OK = "not-tdms"

    return ext_OK

# check that files exist on local file system
def check_files_exist (file_list):

    files_exist = True

    for file in file_list:
        if not os.path.isfile(file):
            files_exist = False

    return files_exist

# get file suffixes for .zip file
def get_suffixes (zip_file):

    # save results as a set
    tdms_suffixes = set()

    # look through zip file for suffixes
    with zipfile.ZipFile(zip_file, 'r') as zid:

        # get zip file content list
        zip_contents = zid.namelist()

        for file in zip_contents:

            # get file name and extension
            head, tail = os.path.split(file)
            ext = tail.split(".")[-1].lower()

            # is it a tdms file?
            if ext == 'tdms' or ext =='tdm':

                # get suffix
                suffix = tail.split("_")[-1].split(".")[0]

                # add suffix to set
                tdms_suffixes.add(suffix)

    return tdms_suffixes

# test that a string converts to an integer
def is_int(s):
    try: 
        int(s)
        return True
    except ValueError:
        return False

# create TDMS model, show progress by default, arguments 
# include the command line connection parameters
# and project/model information
def upload_model (arguments, parser, parms, file_list, progress=True):

    # setup a connection to the Slycat Web Server.
    connection = slycat.web.client.connect(arguments)

    # create a new project to contain our model.
    pid = connection.find_or_create_project(arguments.project_name, arguments.project_description)

    # create the new, empty model.
    mid = connection.post_project_models(pid, "DAC", arguments.model_name, arguments.marking, 
        arguments.model_description)

    # assign model UI defaults
    connection.put_model_parameter(mid, "dac-ui-parms", dac_model_defaults())

    # upload data to new model
    connection.upload_files(mid, file_list, parser, [parms, ["DAC"]], progress)

    # mark model as finished
    connection.post_model_finish(mid)

    # wait until the model is ready
    connection.join_model(mid)

    return mid

# check arguments and create model
def create_model (arguments, log):
    
    # can't have both overvoltage and sprytron
    if arguments.overvoltage and arguments.sprytron:
        raise TDMSUploadError("Can't use both overvoltage and sprytron options " +
              "together. Please select one or the other and try again.")

    # check file name extensions
    file_list = arguments.files
    file_type = check_file_extensions (file_list)
    if file_type == "not-tdms":
        log("One or more files had the wrong extension (note that if using .zip " +
              "only one file can be uploaded). Please revise the file list " +
              "and try again.")
        raise TDMSUploadError("One or more files had the wrong extension " + 
              "(note that if using .zip only one file can be uploaded). Please " + 
              "revise the file list and try again.")

    # check that files exist
    if not check_files_exist (file_list):
        raise TDMSUploadError("One or more input files did not exist. Please make " +
              "sure the file names are correct and try again.")

    # check that zip file is valid
    if file_type == 'zip':
        if not zipfile.is_zipfile(file_list[0]):
            raise TDMSUploadError("Zip file is invalid or corrupt. Please fix the " + 
                  "file and try again.")

    # parser tdms options
    parser_parms = dac_tdms_util.get_parms (arguments)

    # check common parameters
    parser_parms, check_parser_error = dac_tdms_util.check_parms (arguments, parser_parms)
    if check_parser_error != "":
        raise TDMSUploadError(check_parser_error)

    # compile suffixes to include if .zip file
    dac_parser = "dac-tdms-file-parser"
    if file_type == "zip":

        # get all possible suffixes
        include_suffixes = get_suffixes (file_list[0])

        # make sure there is at least one suffix
        if len(include_suffixes) == 0:
            raise TDMSUploadError("No valid .tdms files were found in .zip file. " +
                  "A valid .tdms file has to have a .tdms extension and " +
                  "a recognizable suffix. Please fix .zip file and try again.")

        # exclude suffixes as requested
        if arguments.exclude:

            # exclude and print list of excluded suffixes
            for suffix in arguments.exclude:
                include_suffixes.discard(suffix)

        # sort suffixes
        list_suffixes = list(include_suffixes)
        list_suffixes.sort()

        # add to parameters
        parser_parms.append(list_suffixes)

        # set parser type to zip
        dac_parser = "dac-tdms-zip-file-parser"

    # add file list if not zip file
    else:
        parser_parms.append(file_list)

    # echo back user input, starting with files
    log('*********** Creating DAC Model ***********')
    log("Input file(s):")
    for file in file_list:
        log("\t%s" % file)
    
    # next list included/excluded suffixes
    if file_type == "zip":

        if arguments.exclude:
            log("Excluding TDMS file suffixes:")
            for suffix in arguments.exclude:
                log("\t%s" % suffix)

        log("Including TDMS file suffixes:")
        for suffix in parser_parms[9]:
            log("\t%s" % suffix)

    # next list common parameters
    log("Minimum number of time steps per channel: %s" % parser_parms[0])
    log("Minumum number of channels: %s" % parser_parms[1])
    log("Minimum number of shots: %s" % parser_parms[2])
    log("Number of landmarks: %s" % parser_parms[3])
    log("Number of PCA components: %s" % parser_parms[4])
    log("Expecting TDMS data type: %s" % parser_parms[5])
    log("Combining mismatched time steps using: %s" % parser_parms[6])
    log("Infer channel units: %s" % parser_parms[7])
    log("Infer time units: %s" % parser_parms[8])
    
    # upload model file(s)
    mid = upload_model (arguments, dac_parser, parser_parms, file_list, progress=True)

    # supply the user with a direct link to the new model.
    host = arguments.host
    if arguments.port:
        host = host + ":" + arguments.port
    log("Your new model is located at %s/models/%s" % (host, mid))
    log('***** DAC Model Successfully Created *****')

# logging is just printing to the screen
def log (msg):
    print(msg)

# set up argument parser
def parser ():

    # provide additional command line arguments for TDMS files
    parser = slycat.web.client.ArgumentParser(description=
        "Creates a Dial-A-Cluster model from one or more .tdms files, " +
        "or from a .zip archive containing .tdms files.")

    # actual files to upload
    parser.add_argument("files", nargs='+', 
        help='TDMS file(s) or TDMS .zip files. If ' +
             'file names include spaces, use quotes, ' +
             'e.g. "file name with spaces.tdms".')

    # model and project names/descriptions
    parser.add_argument("--marking", default="cui", 
        help="Marking type.  Default: %(default)s")
    parser.add_argument("--model-description", default="", 
        help="New model description.  Default: %(default)s")
    parser.add_argument("--model-name", default="TDMS DAC Model", 
        help="New model name.  Default: %(default)s")
    parser.add_argument("--project-description", default="", 
        help="New project description.  Default: %(default)s")
    parser.add_argument("--project-name", default="TDMS DAC Models", 
        help="New project name.  Default: %(default)s")

    # add tmds options
    parser = dac_tdms_util.add_options (parser)

    return parser

# tdms entry point
def main():

    # set up argument parser
    tdms_parser = parser()  

    # get user arguments
    arguments = tdms_parser.parse_args()

    # check arguments and create model
    create_model(arguments, log)

# command line version to load a single DAC model
if __name__ == "__main__":

    main()


#!/bin/env python
# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

import pprint
import slycat.web.client

parser = slycat.web.client.ArgumentParser("Get model details from Slycat server.")
parser.add_argument("mid", help="The ID of the model to retrieve")
arguments = parser.parse_args()

connection = slycat.web.client.connect(arguments)
model = connection.get_model(arguments.mid)
pprint.pprint(model)
